/*
 * VDMFec_0005JNI.c (040105) call fec.c of Luigi Rizzo
 * -- forward error correction for Java Native Interface based on Vandermonde matrices
 * 980614
 *
 * (C) 2004
 * Sylvain Dattrino (s-datt02@univ-lyon1.fr)
 * Cong-Dug Pham    (congduc.pham@ens-lyon.fr)
 *
 * (C) 1997-98 Luigi Rizzo (luigi@iet.unipi.it)
 *
 *
 * Portions derived from code by Phil Karn (karn@ka9q.ampr.org),
 * Robert Morelos-Zaragoza (robert@spectra.eng.hawaii.edu) and Hari
 * Thirumoorthy (harit@spectra.eng.hawaii.edu), Aug 1995
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:

 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials
 *    provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 */

#include <malloc.h>

#include "VDMFec_0005fJNI.h"
#include "../fec.c"

// struct fec_parms {
//  u_long magic ;
//  int k, n ;		/* parameters of the code */
//  gf *enc_matrix ;
//} ;

/*
 * gf is the type used to store an element of the Galois Field.
 * Must constain at least GF_BITS bits.
 *
 * Note: unsigned char will work up to GF(256) but int seems to run
 * faster on the Pentium. We use int whenever have to deal with an
 * index, since they are generally faster.
 */

/*
#if (GF_BITS < 2  && GF_BITS >16)
#error "GF_BITS must be 2 .. 16"
#endif
#if (GF_BITS <= 8)
typedef unsigned char gf;
#else
typedef unsigned short gf;
#endif
*/

/******************************************************************************
 *
 * => For each function see java VDMFec_JNI class file for more informations.
 */

JNIEXPORT jint JNICALL Java_VDMFec_1JNI_fec_1new
(JNIEnv * jenv, jclass jclass_jni, jint jk, jint jn)
{
  return ((jint) fec_new((int) jk,(int) jn));
}

JNIEXPORT void JNICALL Java_VDMFec_1JNI_fec_1free
(JNIEnv *jenv, jclass jclass_jni, jint jp)
{
  fec_free((struct fec_parms *)jp);
  return;
}

JNIEXPORT void JNICALL Java_VDMFec_1JNI_init_1fec
(JNIEnv *jenv, jclass jclass_jni)
{
  init_fec();
  
  return;
}

JNIEXPORT void JNICALL Java_VDMFec_1JNI_fec_1encode
(JNIEnv *jenv, jclass jclass_jni, jint jcode, jobjectArray jsrc, jbyteArray jfec, jint jindex, jint jsz)
{
  int i,k=((struct fec_parms *) jcode)->k;

  jbyteArray jPktArray;
  gf *src[k],*fec;

  for( i=0; i<k; i++)
    {
      jPktArray=(jbyteArray)(*jenv)->GetObjectArrayElement(jenv,jsrc, i);

      if(jPktArray)
	src[i]=(gf*) (*jenv)->GetByteArrayElements(jenv,jPktArray, 0);     // cast for jbtye* to gf*
    }

  fec = (gf*) (*jenv)->GetByteArrayElements(jenv,jfec, 0);     // cast for jbtye* to gf*

  fec_encode((struct fec_parms *) jcode,src,fec,(int) jindex,(int) jsz);

  // Commit changes to the Java Virtual Machine

  (*jenv)->ReleaseByteArrayElements(jenv, jfec, fec, 0);

  for(i=0;i<k;i++)
    {

      jPktArray=(jbyteArray)(*jenv)->GetObjectArrayElement(jenv,jsrc, i);

      (*jenv)->ReleaseByteArrayElements(jenv, jPktArray, src[i], 0);

    }

  return;
}

JNIEXPORT jint JNICALL Java_VDMFec_1JNI_fec_1decode
(JNIEnv *jenv, jclass jclass_jni, jint jcode, jobjectArray jpkt, jintArray jindex, jint jsz)
{
  int i,k=((struct fec_parms *) jcode)->k;
  
  jbyteArray jPktArray;
  gf *pkt[k];
  int *index,retval;

  int sz=(int)jsz;

  for( i=0; i<k; i++)
    {
      jPktArray=(jbyteArray)(*jenv)->GetObjectArrayElement(jenv,jpkt, i);

      if(jPktArray)
	pkt[i]=(gf*) (*jenv)->GetByteArrayElements(jenv,jPktArray, 0);     // cast for jbtye* to gf*
    }

  index = (int*) (*jenv)->GetIntArrayElements(jenv,jindex, 0);     // cast for jbtye* to int*

  retval=fec_decode(((struct fec_parms *) jcode),pkt,index,sz);

  // Commit changes to the Java Virtual Machine

  for(i=0;i<k;i++)
      {

	jPktArray=(jbyteArray)(*jenv)->GetObjectArrayElement(jenv,jpkt, i);

	(*jenv)->ReleaseByteArrayElements(jenv, jPktArray, pkt[i], 0);

      }

  (*jenv)->ReleaseIntArrayElements(jenv, jindex,(jint *) index, 0);

  return retval;
}

JNIEXPORT jint JNICALL Java_VDMFec_1JNI_test_1decode
(JNIEnv *jenv, jclass jclass_jni, jint jcode, jint jk, jintArray jindex, jint jsz, jstring js)
{
  jint retval;

  int * index = (int *) (*jenv)->GetIntArrayElements(jenv,jindex, JNI_FALSE);

  char * s = (char*) (*jenv)->GetStringUTFChars(jenv, js, JNI_FALSE);

  retval = test_decode((void *)jcode, (int) jk, index, (int) jsz, s);

  // Commit changes to the Java Virtual Machine

  (*jenv)->SetIntArrayRegion(jenv, jindex, 0, (int) jk, (jint *)index);

  (*jenv)->ReleaseStringChars(jenv, js, (const jchar*) s);

  return retval;

}
