/*
 * VDMFec_JNI.java (040105) based on fec.c of Luigi Rizzo
 * -- forward error correction for Java Native Interface based on Vandermonde matrices
 * 980614
 *
 * (C) 2004
 * Sylvain Dattrino (s-datt02@univ-lyon1.fr)
 * Cong-Dug Pham    (congduc.pham@ens-lyon.fr)
 *
 * (C) 1997-98 Luigi Rizzo (luigi@iet.unipi.it)
 *
 *
 * Portions derived from code by Phil Karn (karn@ka9q.ampr.org),
 * Robert Morelos-Zaragoza (robert@spectra.eng.hawaii.edu) and Hari
 * Thirumoorthy (harit@spectra.eng.hawaii.edu), Aug 1995
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:

 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials
 *    provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 */

import java.util.*;
import java.text.*;

/**
 * This is an erasure code based on Vandermonde matrices class.
 * <p>
 * WARNING: the following class can be much more lower than native implementation due to the
 * Java Virtual Machine, furthermore this is only encapsulation of native code, hence portability of original 
 * code is NOT improve.
 */

public class VDMFec_JNI
{

    /**
     * The following parameter defines how many bits are used for
     * field elements. The code supports any value from 2 to 16
     * but fastest operation is achieved with 8 bit elements
     * This is the only parameter you may want to change.
     */

    public final static int GF_BITS = 8;     /* code over GF(2**GF_BITS) - change to suit */

    private final static int GF_SIZE = (1 << GF_BITS) - 1;     /* powers of \alpha */

    /**
     * Number of maximum source packets
     * <p>
     * Must be inferior to 255, use in main test programme
     */

    public final static int KK = 64;     /* 255 */

    /**
     * Size of packet
     * <p>
     * Size in bytes of each packet to generate
     */

    public final static int SZ = 1024;

    private static int prev_k = 0, prev_sz = 0;

    private static byte[][] d_original = null, d_src = null ;

    // Library of native code to load.
    
    static {System.loadLibrary("vdm_fec");}

    /**
     * TICK : Take a timestamp
     * <p>
     * @param t        (IN) array of reference times
     * @param index    (IN) position in t where to put the new reference
     *
     */

    public static void TICK(long[] t,int index)
    {

	Date t1;

	try
	    {
		t1= new Date();
		
		t[index]=(t1.getTime()*1000);
	    }
	catch(OutOfMemoryError e)
	    {
		System.out.print("Error: insufficient memory (new failed for ticks)\n");
		
		System.exit(-1);
	    }

	return;

    }

    /**
     * TOCK : Compute a time interval
     * <p>
     * @param t        (IN) array of reference times
     * @param index    (IN) position in t where old reference is and where new reference will be put
     *
     */

    public static void TOCK(long[] t,int index)
    {
	Date t1;

	try
	    {
		t1 = new Date();
		
		if((t[index]=(t1.getTime()*1000)-t[index])==0)
		    t[index]=1;
	    }

	catch(OutOfMemoryError e)
	    {
		System.out.print("Error: insufficient memory (new failed for ticks)\n");
		
		System.exit(-1);
	    }

	return;
    }

    /**
     * fec_new : create a new encoder for a fec session
     * <p>
     * This contains k,n and the encoding matrix.
     * <p>
     * @return     descriptor of the (struc fec_parms *) native type
     */

    public static native int fec_new(int k,int n);

    /**
     * fec_free : free some space
     * <p>
     * Function allowing to free memory allocated in native environement
     *
     * @param p     (IN) descriptor of the (struc fec_parms *) native type
     *
     */

    public static native void fec_free(int p);

    /**
     * init_fec : initialize fec sessions
     *
     */

    public static native void init_fec();

    /**
     * fec_encode : encode one fec packet
     * <p>
     * @param code     (IN)  descriptor of the (struc fec_parms *) native type
     * @param src      (IN)  pointer to n data packets
     * @param dst      (OUT) pointer to computed output packet
     * @param index    (IN)  value used for computation
     * @param sz       (IN) size of each packets
     */

    public static native void fec_encode(int code, byte[][] src, byte[] fec, int index, int sz) ;

    /**
     * fec_decode : decode fec packets of the session
     * <p>
     * @param code     (IN)     pointer to code descriptor
     * @param pkt      (IN-OUT) pointers to received packets
     * @param index    (IN)     values used for computation
     * @param sz       (IN)     size of each packet
     * <p>
     * @return correct vector
     */

    public static native int fec_decode(int code, byte[][] pkt, int[] index, int sz) ;

    /**
     * test_decode : test function of VDM Library
     * <p>
     * the following is only test code and can be safely omitted
     * in applications.
     * Creates k packets of size sz of random data, encodes them,
     * and tries to decode.
     * Index contains the permutation entry.
     *
     * @param code     (IN)     pointer to code descriptor
     * @param k        (IN)     number of maximum source packets
     * @param index    (IN)     value used for computation
     * @param sz       (IN)     size of each packet
     * @param s        (IN)     description of the test
     * <p>
     * @return         number of errors
     */

    public static int test_decode(int code, int k, int[] index, int sz, String s)
    {
	int errors;
	int reconstruct = 0 ;
	int item, i ;

	long[] ticks;

	try
	    {
		ticks = new long[10];
	    }
	catch(OutOfMemoryError e)
	    {
		System.out.print("Error: insufficient memory (new failed for ticks)\n");
		
		return -1;
	    }

	if (sz < 1 || sz > 8192) {
	    System.err.print("test_decode: size "+sz+" invalid, must be 1..8K\n");
	    return 1 ;
	}

	if (k < 1 || k > GF_SIZE + 1) {
	    System.err.print("test_decode: k "+k+" invalid, must be 1.."+(GF_SIZE + 1)+"\n");
	return 2 ;
	}

	if (prev_k != k || prev_sz != sz) {
	    if (d_original != null) {
		for (i = 0 ; i < prev_k ; i++ ) {
		    d_original[i]=null;
		    d_src[i]=null;
		}

		d_original=null;
		d_src=null;
	    }
	}

	prev_k = k ;
	prev_sz = sz ;

	if (d_original == null) {

	    try
		{
		    d_original = new byte[k][sz];
		    d_src = new byte[k][sz];
		}
	    catch(OutOfMemoryError e)
		{
		    System.out.print("Error: insufficient memory (new failed for d_original or d_src)\n");
		    
		    d_original = null;

		    return -1;
		}

	    /*
	     * build sample data
	     */

	    for (i = 0 ; i < k ; i++ ) {
		for (item=0; item < sz; item++)
		    d_original[i][item] = (byte) ( ((item ^ i) + 3) & GF_SIZE);
	    }
	}

	errors = 0 ;

	for( i = 0 ; i < k ; i++ )
	    if (index[i] >= k ) reconstruct ++ ;

	TICK(ticks,2);

	for( i = 0 ; i < k ; i++ )
	    fec_encode(code, d_original, d_src[i], index[i], sz );
	TOCK(ticks,2);

	TICK(ticks,1);
	if (fec_decode(code, d_src, index, sz) != 0) {
	    System.err.print("detected singular matrix for "+s+"  \n");
	    
	    return 1 ;
	}

	TOCK(ticks,1);

	for (i=0; i<k; i++)
	    if (!Arrays.equals(d_original[i], d_src[i])) {
		errors++;
		System.err.print("error reconstructing block "+i+"\n");
	    }

	if (errors != 0)
	    System.err.print("Errors reconstructing "+errors+" blocks out of "+k+"\n");

	DecimalFormat OutputFormat_1 = new DecimalFormat("000.000000");

	// DecimalFormat OutputFormat_1 = new DecimalFormat("  0.0  ");

	DecimalFormat OutputFormat_2 = new DecimalFormat("00");

	// DecimalFormat OutputFormat_2 = new DecimalFormat("  0");
	
	System.err.print("  k "+OutputFormat_2.format(k,new StringBuffer(),new FieldPosition(NumberFormat.INTEGER_FIELD))+", l "+OutputFormat_2.format(reconstruct,new StringBuffer(),new FieldPosition(NumberFormat.INTEGER_FIELD))+"  c_enc "+ OutputFormat_1.format((double)(k * sz * reconstruct)/(double)ticks[2],new StringBuffer(),new FieldPosition(NumberFormat.INTEGER_FIELD)) +" MB/s c_dec "+ OutputFormat_1.format((double)(k * sz * reconstruct)/(double)ticks[1])+" MB/s     \r");

	return errors ;
    }

    /**
     * main : main function, launched from Java Virtual Machine
     * <p>
     * Create k packets of size sz of random data, encodes them,
     * and tries to decode.
     * Index contains the permutation entry.
     *
     * @param args     (IN)     arguments of function (not used)
     *
     */

    public static void main(String[] args)
    {

	byte[] buf;

	try
	    {
		buf= new byte[256];
	    }
	catch(OutOfMemoryError e)
	    {
		System.out.print("Error: insufficient memory (new failed for packetsArray)\n");
		
		return;
	    }

	int code,kk,i;

	int[] ixs;

	int lim = GF_SIZE + 1 ;

	if (lim > 1024) lim = 1024 ;

	for ( kk = KK ; kk > 2 ; kk-- ) {
	    code = fec_new(kk, lim);

	    try
		{
		    ixs = new int[kk];
		}
	    catch(OutOfMemoryError e)
		{
		    System.out.print("Error: insufficient memory (new failed for packetsArray)\n");

		    buf=null;

		    fec_free(code);

		    return;
		}

	    for (i=0; i<kk; i++) ixs[i] = kk - i ;

	    test_decode(code, kk, ixs, SZ, "kk="+kk+", kk-i");
	
	    for (i=0; i<kk; i++) ixs[i] = i ;
	    test_decode(code, kk, ixs, SZ, "i");
	    
	    if (false) {
		for (i=0; i<kk; i++) ixs[i] = i ;
		ixs[0] = ixs[kk/2] ;
		test_decode(code, kk, ixs, SZ, "0 = 1 (error expected)");
	    }
	    
	    if (false)
		for (i= lim-1 ; i >= kk ; i--) {
		    int j ;
		    for (j=0; j<KK; j++) ixs[j] = kk - j ;
		    ixs[0] = i ;
		    test_decode(code, kk, ixs, SZ, "0 = big");
		}
	    
	    if (false)
		for (i= lim - kk ; i >= 0 && i>= lim - kk - 4 ; i--) {
		    int j ;
		    for (j=0; j<kk; j++)
			ixs[j] = kk -1 - j + i ;
		    test_decode(code, kk, ixs, SZ, "shifted j");
		}
	    
	    if (true)  {
		int j, max_i0 = KK/2 ;
		if (max_i0 + KK > lim)
		    max_i0 = lim - KK ;
		for (i= 0 ; i <= max_i0 ; i++) {
		    for (j=0; j<kk; j++)
			ixs[j] = j + i ;
		    test_decode(code, kk, ixs, SZ, "shifted j");
		}
	    }

	    System.err.println("");
	    ixs=null;
	    fec_free(code);
	}

	return;
    }
}
